unit UnitClipQueue;
{
    Purpose:
        Encapsulate all the rules of of the queue
        - number of items allowed
        - deleting icons on item removal (no memory leaks)
        - etc
        Handle a separate queue of items removed from the ClipQueue

    Updates:

        Created a basic StringQueue that's just a FIFO to be
        used for a RemovedQueue object.
        MoveDuplicateTop logic moved here (where it should be)
        Ignore duplicates durring Add or Instert
}

interface
uses classes, Windows;

type TStringQueue = class(TObject)
    private
        sl : TStringList;
        qSize : cardinal;
    public
        constructor Create;
        destructor Destroy; override;
        procedure SetQueueSize(size : longint);
        procedure DeleteQueueItem(index : cardinal); virtual;
        procedure InsertAtStart(s : string);

        function GetQueueCount : cardinal;
        function GetQueueItem(index : cardinal) : string;
end;

type TClipQueue = class(TStringQueue)
    private
        MoveDuplicateTop : boolean;
    public
        constructor Create;
        destructor Destroy; override;

        {config options}
        procedure SetMoveDuplicateTop(enable: boolean);

        {}

        function GetQueueIcon(index : cardinal) : thandle;
        procedure SetQueueIcon(index : cardinal; h : THandle);

        procedure MoveToStart(index : cardinal);
        procedure Move(oldIndex : cardinal; newIndex : cardinal);

        procedure Add(s : string);
        procedure InsertAtStart(s : string; h : HICON);

        function IndexOf(s : string) : cardinal;
        procedure DeleteQueueItem(index : cardinal); override;
        procedure ClearQueue;
        procedure GetQueueItems(items : TStrings);
        procedure SetQueueItems(items : TStrings);
end;

var ClipQueue : TClipQueue;
var RemovedQueue : TStringQueue;

implementation

uses UnitFrmMainPopup;



//////////////////////////
//
// TStringQueue
// - Basic FIFO for strings
// - Enforces a total size
////////////////////////////

//-------------------
// (Con/De)structors
//-------------------
constructor TStringQueue.Create;
begin
    sl := TStringList.Create;
end;

destructor TStringQueue.Destroy;
begin
    sl.free;
    inherited Destroy;
end;

procedure TStringQueue.SetQueueSize(size : longint);
begin
    qSize := size;
    while (sl.count > size) and (sl.count > 0) do
        self.DeleteQueueItem(sl.count - 1);
end;

procedure TStringQueue.DeleteQueueItem(index : cardinal);
begin
    sl.Delete(index);
end;

procedure TStringQueue.InsertAtStart(s : string);
begin
    sl.Insert(0,s);
    self.SetQueueSize(self.qSize); 
end;


function TStringQueue.GetQueueCount : cardinal;
begin
    result := sl.count;
end;

function TStringQueue.GetQueueItem(index : cardinal) : string;
begin
    result := sl[index];
end;


//////////////////////////
//
// TClipQueue
//
////////////////////////////

//-------------------
// (Con/De)structors
//-------------------
constructor TClipQueue.Create;
begin
    sl := TStringList.Create;
end;

destructor TClipQueue.Destroy;
begin
    sl.free;
    inherited Destroy;
end;


//-------------------
// Configuration
//-------------------
procedure TClipQueue.SetMoveDuplicateTop(enable: boolean);
begin
    self.MoveDuplicateTop := enable;
end;

{
procedure TClipQueue.SetQueueSize(size : longint);
begin
    qSize := size;
    while (sl.count > size) and (sl.count > 0) do
        self.DeleteQueueItem(sl.count - 1);
end;
}

//-------------------
// Public Interface
//-------------------


//
// items added to list
//
procedure TClipQueue.Add(s : string);
var i : longint;
begin
    i := sl.IndexOf(s);
    if (i = -1) then begin
        sl.Add(s);
    end else if (self.MoveDuplicateTop) then begin
        self.MoveToStart(i);
    end;
end;

procedure TClipQueue.InsertAtStart(s : string; h : HICON);
var i : longint;
begin
    // Unique items only
    i := sl.IndexOf(s);
    if (i = -1) then begin
        sl.InsertObject(0,s, TObject(h));
        self.SetQueueSize(self.qSize);
    end else if (self.MoveDuplicateTop) then begin
        self.MoveToStart(i);
    end;
end;

procedure TClipQueue.SetQueueItems(items : TStrings);
begin
    //
    // trim queue if needed

    self.ClearQueue;
    sl.AddStrings(items);
    self.SetQueueSize(self.qSize);
end;



function TClipQueue.IndexOf(s : string) : cardinal;
begin
    result := sl.IndexOf(s);
end;


procedure TClipQueue.GetQueueItems(items : TStrings);
begin
    items.Clear;
    items.AddStrings( sl );
end;


function TClipQueue.GetQueueIcon(index : cardinal) : thandle;
begin
    result := THandle(sl.Objects[index]);
end;


procedure TClipQueue.SetQueueIcon(index : cardinal; h : THandle);
begin
    sl.objects[index] := TObject(h);
end;




//
// Move items in the list
//
procedure TClipQueue.MoveToStart(index : cardinal);
begin
    sl.Move(index, 0);
end;

procedure TClipQueue.Move(oldIndex : cardinal; newIndex : cardinal);
begin
    sl.Move(oldIndex, newindex);
end;



//
// Remove items from queue
// NOTE: Resources must be released here
//
procedure TClipQueue.DeleteQueueItem(index : cardinal);
begin
    // We've got to clean up the clone icon to avoid a memory leak
    if (sl.Objects[index] <> nil) then begin
        frmMainPopup.AppendLog('^Deleting Icon^');
        Windows.DestroyIcon(HICON(sl.Objects[index]));
    end;

    // this MUST be the ONLY place an item is deleted
    RemovedQueue.InsertAtStart(self.GetQueueItem(index));
    sl.Delete(index);
end;


procedure TClipQueue.ClearQueue;
var i : longint;
begin
    // see note in DeleteQueueItem for rules of removing items
    // from queue
    // I know better, this must go in reverse order
    for i := (sl.count - 1) downto 0 do begin
        self.DeleteQueueItem(i);
    end;

end;




//
// Static Objects
//
initialization
    RemovedQueue := TStringQueue.Create;
    ClipQueue := TClipQueue.Create;
end.
