unit UnitFrmClipboardManager;
{
    Purpose:
    This unit monitors new items on the clipboard.
    Raw data is retrieved and send to the ClipQueue
    (the text, filenames, and icons).

    Also, this unit is the interface for the user
    to manipulating the queue.

    NOTE:
    Items entering the clipboard too soon after an ArsClip
    paste are ignored, since our clipboard items must be
    ignored.

Updates:
    Improved error reporting for corrupt data file

    Two Icon notifications when an item is copied: normal
    mode, and disabled mode

    --

    A new method of ignoring the clipboard was needed when another
    program was part of the clipboard viewer chain. When a
    WMDRAWCLIPBOARD message arrives, at least 1 second must have
    passed since the last time FrmMainPopup placed an item on the
    clipboard.

    The SetIgnoreClipboard() function is only needed durring
    a paste operation.

    Moved the rest of the Queue logic to UnitClipQueue

    Copied files no longer causes duplicate entries

    lbQueue is not longer used for the clipboard history, due to Win9X
    memory constraints for components. See UnitClipQueue for details.

}
interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, Clipbrd, ShellAPI, Buttons, ExtCtrls, ComCtrls {for DragQueryFile};

type
  TfrmClipboardManager = class(TForm)
    lbQueue: TListBox;
    Panel1: TPanel;
    btnTop: TSpeedButton;
    btnUp: TSpeedButton;
    btnDown: TSpeedButton;
    btnBottom: TSpeedButton;
    Panel2: TPanel;
    btnClose: TButton;
    btnDelete: TButton;
    mPreview: TMemo;
    Preview: TLabel;
    bCopyToClipboard: TButton;
        procedure FormCreate(Sender: TObject);
        procedure FormDestroy(Sender: TObject);
        procedure btnCloseClick(Sender: TObject);
        procedure btnTopClick(Sender: TObject);
        procedure btnUpClick(Sender: TObject);
        procedure btnDownClick(Sender: TObject);
        procedure btnBottomClick(Sender: TObject);
        procedure btnDeleteClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormShow(Sender: TObject);
    procedure lbQueueClick(Sender: TObject);
    procedure bCopyToClipboardClick(Sender: TObject);
    private

        CF_HTML : UINT;
        CF_RICHTEXT : UINT;
        HistoryLoaded : boolean;
        { Private declarations }
        NextHandle: THandle;                {clipboard chain pointer}
        DisableMonitoring: boolean;         {altered by public interface}
        IgnoreClipboard: boolean;
        MonitorFilenames: boolean;
        MonitorOtherItems: boolean;
        CopyIcon : boolean;
        LastPaste : TDateTime;
        DisableWhenScrollLock : boolean;

        procedure ClipboardChainJoin;
        procedure ClipboardChainLeave;

        {windows messages}
        procedure WMDRAWCLIPBOARD(var Message: TMessage); message WM_DRAWCLIPBOARD;
        procedure WMCHANGECBCHAIN(var Message: TMessage); message WM_CHANGECBCHAIN;

    public
        { Public declarations }

        {configuration}
        procedure SetMonitoring(disable: boolean);
        procedure SetFilenameMonitoring(enable: boolean);
        procedure SetOtherItemsMonitoring(enable: boolean);
        procedure SetCopyIcon(enable: boolean);
        procedure SetDisableWhenScrollLock(value : boolean);

        {temporarily ignore the clipboard items}
        procedure SetIgnoreClipboard(ignore : boolean);
        function GetClipboardOwnerIcon : HICON;

        {history load/save}
        //function GetDoSave
        procedure LoadHistory;
        procedure SaveHistory;


        procedure InformOfPaste;
        procedure RefreshClipboardMonitor;
  end;

var
  frmClipboardManager: TfrmClipboardManager;

implementation

uses UnitFrmMainPopup, UnitListBoxMover, UnitFrmDummyIcon,
    UnitClipQueue, UnitOtherQueue, UnitKeyboardQuery, StrUtils, DateUtils,
  UnitFrmSysTrayMenu, UnitFrmChainWatcher;
const REMEMBER_FILE = 'remember.txt';


{$R *.dfm}


{
--======================
-- // Public Inteface //
--======================
}
procedure TfrmClipboardManager.RefreshClipboardMonitor;
begin
    self.ClipboardChainLeave;
    self.ClipboardChainJoin; 
end;
procedure TFrmClipboardManager.InformOfPaste;
begin
    LastPaste := SysUtils.Time();
end;

//--------------------------------
// configuration stuff
//--------------------------------

procedure TfrmClipboardManager.SetCopyIcon(enable: boolean);
begin
    self.CopyIcon := enable;
end;
procedure TfrmClipboardManager.SetMonitoring(disable: boolean);
begin
    self.DisableMonitoring := disable;
end;

procedure TfrmClipboardManager.SetIgnoreClipboard(ignore : boolean);
begin
    IgnoreClipboard := ignore;
    frmMainPopup.AppendLog('ClipboardManager: ignore = ' + BoolToStr(ignore) );
end;

procedure TfrmClipboardManager.SetFilenameMonitoring(enable: boolean);
begin
    MonitorFilenames := enable;
end;

procedure TFrmClipboardManager.SetOtherItemsMonitoring(enable: boolean);
begin
    MonitorOtherItems := enable;
    if (not enabled) then
        ClipQueue.ClearQueue;
end;

procedure TFrmClipboardManager.SetDisableWhenScrollLock(value : boolean);
begin
    DisableWhenScrollLock := value;
end;

{
--======================
-- // Create/Destroy  //
--======================
}


procedure TfrmClipboardManager.FormCreate(Sender: TObject);
begin


   self.ClipboardChainJoin;
end;

procedure TfrmClipboardManager.FormDestroy(Sender: TObject);
begin
  self.ClipboardChainLeave;
end;




{
--=========================
-- // Clipboard messages //
--=========================

Description: Monitor for new text entries entered into the clipboard. Add
new text entries into the finite sized queue (removing oldest items as needed).
Also, follow the rule for a change in the clipboard chain.

}
procedure TfrmClipboardManager.WMDRAWCLIPBOARD(var Message: TMessage);
var s : string;
    cTime : TDateTime;
label exit_code;

    function GetClipboardText: string;
    var hnd : THandle;
        p : Pointer;

        format : UINT;
        formatFound : boolean;
    begin
        // This code is an attemp to get rid of the "Cannot Open Clipboard"
        // errors. The previous version of this code caused garbage to appear
        // in the history.
        result := '';
        p := nil;
        hnd := 0;
        try
            formatFound := false;
            format := 0;
            {
            //
            // possible future enhancements
            //
            if (Clipboard.HasFormat(CF_RICHTEXT)) then begin
                format := CF_RICHTEXT;
                formatFound := true;
            end else if (Clipboard.HasFormat(CF_HTML)) then begin
                format := CF_HTML;
                formatFound := true;
            end else if (clipboard.HasFormat(CF_TEXT)) then begin
                format := CF_TEXT;
                formatFound := true;
            end;
            }
            if (clipboard.HasFormat(CF_TEXT)) then begin
                format := CF_TEXT;
                formatFound := true;
            end;
            if (formatFound) then begin
                frmMainPopup.AppendLog('open clipboard');
                if (Windows.OpenClipboard(Application.Handle)) then begin
                    hnd := Windows.GetClipboardData(format);
                    if (hnd <> 0)  then begin
                        p := GlobalLock(hnd);
                        if (p <> nil) then begin
                            result := PChar(p);
                        end else begin
                            frmMainPopup.AppendLog('ERROR: failed to lock clipboard data');
                        end;
                    end else begin
                        frmMainPopup.AppendLog('ERROR: GetClipboardData() : ' + SysErrorMessage(GetLastError()) );
                    end;
                    Windows.CloseClipboard;
                end else begin
                    frmMainPopup.AppendLog('ERROR: Failed to open clipboard: ' + SysErrorMessage(GetLastError()) );
                end;
            end;
        finally
            if (p <> nil) and (hnd <> 0) then begin
                Windows.GlobalUnlock(hnd);
                frmMainPopup.AppendLog('closing lock');
            end else begin
                frmMainPopup.AppendLog('not closing lock');
            end;
        end;

    end;

    //
    // return all filenames or return empty string
    function GetFilenames: string;
    var h, lockdata : THandle;
        i, j : longint;
        s : string;
    begin
        result := '';

        if (MonitorFilenames) then begin
            if (clipboard.HasFormat(CF_HDROP)) then begin
                frmMainPopup.AppendLog('    Detecting Filenames...');

                // only attemp after a global lock has been made
                h := clipboard.GetAsHandle(CF_HDROP);
                lockdata := THandle(Windows.GlobalLock(h));
                if (lockdata <> 0) then begin
                    j := ShellApi.DragQueryFile(lockdata, $FFFFFFFF, nil, 0);

                    result := '';
                    for i := 0 to (j - 1) do begin
                        s := stringofchar(' ', Windows.MAX_PATH);
                        ShellApi.DragQueryFile(lockdata, i, pchar(s), length(s));
                        frmMainPopup.AppendLog(Trim(String(s)));
                        if i = (j - 1) then begin
                            result := result + Trim(PChar(s));
                        end else begin
                            result := result + Trim(PChar(s)) + #13 + #10;
                        end;
                    end;

                    Windows.GlobalUnlock(lockdata);
                end;
            end;
        end;
    end;
begin
    if (FrmChainWatcher <> nil) then begin
        FrmChainWatcher.NotifyOfClipboardActivity;
    end;
    
    // is this our data that we just pasted?
    // disabled? Ignore the clipboard for now?
    // make sure the data isn't from us
    // configurable disable via Scroll Lock key

    CTime := SysUtils.Time();
    if ((MillisecondOfTheDay(CTime) - MillisecondOfTheDay(LastPaste)) < 300) then begin
        frmMainPopup.AppendLog('    Too soon a paste message - ignoring');
        goto exit_code;
    end;

    if (DisableMonitoring) then begin
        frmSysTrayMenu.SetIconIgnore;
        frmMainPopup.AppendLog('//// Monitoring disabled');
        goto exit_code;
    end;
    if (IgnoreClipboard) then begin
        frmMainPopup.AppendLog('//// Ignoring ');
        goto exit_code;
    end;

    if (Windows.GetClipboardOwner = self.Handle) then begin
        frmMainPopup.AppendLog('    Ignore our own clipboard events');
        goto exit_code;
    end;

    If (self.DisableWhenScrollLock) then begin
        if (KeyboardQuery.LockOn(VK_SCROLL)) then begin
            frmSysTrayMenu.SetIconIgnore;

            frmMainPopup.AppendLog('    scroll lock on - ignoring contents');
            goto exit_code;
        end;
    end;

    // get ready to do the thing
    IgnoreClipboard := true;
    frmMainPopup.AppendLog('//// Clipboard Start ////' +
        #13+#10 +
        '    # = ' + IntToStr(message.msg) +
        ' lparam = ' + IntToStr(message.LParam) +
        ' wparam = ' + IntToStr(message.WParam)
    );


    frmSysTrayMenu.SetIconTriggered;
    //
    // Clipboard contents have changed
    // Copy the contents if it is text
    try
        frmMainPopup.AppendLog('    Attempting to retrieve clipboard text');
        s := GetClipboardText;

        if (s = '') then begin
            // check for filenames
            frmMainPopup.AppendLog('    Empty string in clipboard or error retreiving text.');
            s := GetFilenames();
            if (s = '') and (self.MonitorOtherItems) then begin
                if (CopyIcon) then begin
                    OtherQueue.ClipboardSave(self.GetClipboardOwnerIcon);
                end else begin
                    OtherQueue.ClipboardSave(HICON(nil));
                end;
            end;
        end;

        // trim the queue size, dup oldest entry
        if (s <> '') then begin
            if (CopyIcon) then begin
                ClipQueue.InsertAtStart(s, self.GetClipboardOwnerIcon);
            end else begin
                ClipQueue.InsertAtStart(s, HICON(nil));
            end;
        end;
    finally

    end;


    IgnoreClipboard := false;
exit_code:
    // Pass message to next in line (rules of a cliboard viewer)
    Message.Result := 0;
    If (NextHandle <> 0) then begin
        Windows.SendMessage(NextHandle, WM_DRAWCLIPBOARD,  message.WParam, message.LParam )
    end;
    frmMainPopup.AppendLog('//// Clipboard End ////');
end;

procedure TfrmClipboardManager.WMCHANGECBCHAIN(var Message: TMessage);
begin
    //
    //Someone is leaving the chain
    //

    { MS Documentation...
    When a clipboard viewer window receives the WM_CHANGECBCHAIN message,
    it should call the SendMessage function to pass the message to the
    next window in the chain, unless the next window is the window being
    removed. In this case, the clipboard viewer should save the handle
    specified by the lParam parameter as the next window in the chain.
    }
    if (Cardinal(Message.WParam) = NextHandle) then begin
        frmMainPopup.AppendLog('WM_Changecbchain - reassign next handle');
        NextHandle := Message.LParam;
    end else if (NextHandle <> 0) then begin
        frmMainPopup.AppendLog('WM_Changecbchain - send message to next handle');
        if NextHandle <> 0 then
            sendmessage(NextHandle,
                        WM_CHANGECBCHAIN,
                        Message.WParam,  // handle of window to remove
                        Message.LParam); // handle of next window
    end;

    //
    // viewers joining/leaving the chain can send erounious 'clipboard
    // changed' events, so they'll need to be ingored just like our own
    // paste events
    self.InformOfPaste;
    Message.Result := 0;
end;

procedure TfrmClipboardManager.ClipboardChainJoin;
begin
    //Join the cliboard chain
    // Disable monitoring until this form has been created
    // This prevents aditions until the Config form has a
    // chance to set the queue size
    self.SetIgnoreClipboard(true);
    self.CF_HTML := RegisterClipboardFormat('HTML Format');
    self.CF_RICHTEXT := RegisterClipboardFormat('Rich Text Format');
    NextHandle := Windows.SetClipboardViewer(self.Handle);

    self.SetIgnoreClipboard(false);
end;
procedure TfrmClipboardManager.ClipboardChainLeave;
begin
    //Leave the chain
    Windows.ChangeClipboardChain(Handle,     // our handle to remove
                       NextHandle ); // handle of next window in the chain
end;









procedure TfrmClipboardManager.LoadHistory;
var name, itemText, s : string;
    lineCount : cardinal;
    i : integer;
    tf : textfile;
begin
    // only load once
    if (HistoryLoaded) then begin
        exit;
    end;
    HistoryLoaded := true;
    //
    // load permanent items
    //
    name := IncludeTrailingPathDelimiter(ExtractFilePath(application.ExeName)) + REMEMBER_FILE;
    if FileExists(name) then begin
        AssignFile(tf, name);
        Reset(tf, name);

        ClipQueue.ClearQueue;

        //
        // abort reading and show message on error
        // always close the file
        //
        try
            while not eof(tf) do begin
                try
                    Readln(tf, s);
                    itemText := '';
                    lineCount := StrToInt(s);

                    for i := 0 to lineCount - 1 do begin
                        Readln(tf, s);
                        if (itemText = '') then begin
                            itemText := s;
                        end else begin
                            itemText := itemText + #13#10 + s;
                        end;
                    end;

                    ClipQueue.Add(itemText);
                except
                    on E: Exception do begin
                        ShowMessage('The "Load items from last session" file is corrupted - ' + name + #13#10#13#10 +
                                    'Error Message: ' + E.Message);
                        break;
                    end;
                end;
            end;
        finally
            CloseFile(tf);
        end;
    end;
end;

procedure TfrmClipboardManager.SaveHistory;
var name, s : string;
    i, j, cnt : integer;
    tf: textfile;
begin
    //
    // save items
    //
    name := IncludeTrailingPathDelimiter(ExtractFilePath(application.ExeName)) + REMEMBER_FILE;

    AssignFile(tf, name);
    Rewrite(tf);

    for i := 0 to (ClipQueue.GetQueueCount - 1) do begin
        // ultra cheap trick[tm] to count lines in an item
        // FIX: Count CR+LF instead of just CR
        cnt := 1;
        s := ClipQueue.GetQueueItem(i);
        for j := 1 to length(s) - 1 do begin
            if (s[j] = #13) and (s[j+1]= #10) then inc(cnt);
        end;

        writeln(tf, cnt);
        writeln(tf, s);
    end;


    CloseFile(tf);
end;




function TFrmClipboardManager.GetClipboardOwnerIcon() : HICON;
var h, owner : THandle;


    // NOTE: DestroyIcon must before deleting a queued item!!!!!
    function CloneIcon(inIcon : HICON) : HICON;
    var hIconNew : HICON;
        info : _ICONINFO;
    begin
        result := 0;

        // info from PCMagazine's TrayManager
        // Get information about the specified icon
        // Create a clone of the icon

        if (not Windows.GetIconInfo(inIcon, info)) then begin
            frmMainPopup.AppendLog('clone icon: ' + SysErrorMessage(GetLastError));
            EXIT;
        end;

        hIconNew := Windows.CreateIconIndirect(info);
        if (hIconNew = 0) then begin
            frmMainPopup.AppendLog('CreateIconIndirect failed: ' + SysErrorMessage(GetLastError));
        end;

        // Delete the info item's bitmaps -- otherwise we get a very
        // ugly memory leak
        if (not Windows.DeleteObject(info.hbmMask)) then begin
            frmMainPopup.AppendLog('DeleteObject failed: ' + SysErrorMessage(GetLastError));
        end;
        if (not Windows.DeleteObject(info.hbmColor)) then begin
            frmMainPopup.AppendLog('DeleteObject failed: ' + SysErrorMessage(GetLastError));
        end;
        result := hIconNew;

    end;

begin
    // set defualt result
    // bail out if owner went bye-bye
    result := 0;
    owner := Windows.GetForegroundWindow;
    if (owner = 0) then begin
        EXIT;
    end;

    frmMainPopup.AppendLog('^Cloning Icon^');
    // find topmost parent
    repeat
        h := GetWindowLong(owner, GWL_HWNDPARENT);
        if (h <> 0) then begin
            {
            // may be used in the future to identifiy CLIPBRDWNDCLASS

            s := StringOfChar(#0, 255);
            if GetClassName(owner, pchar(s), length(s)-1) <> 0 then begin
                // CLIPBRDWNDCLASS will return nothing
                // wonder what that bastard class actually is
            end;
            }
            owner := h;
        end;
    until (h = 0);
    Windows.SetLastError(ERROR_SUCCESS);

    // get the first small icon available, if any
    h := Windows.GetClassLong(owner, GCL_HICONSM);
    if (h <> 0) then begin
        result := CloneIcon(h);
        EXIT;
    end;
    Windows.SetLastError(ERROR_SUCCESS);

    h := Windows.GetClassLong(owner, GCL_HICON);
    if (h <> 0) then begin
        result := CloneIcon(h);
        EXIT;
    end;
    Windows.SetLastError(ERROR_SUCCESS);

    if (SendMessageTimeout(owner, WM_GETICON, ICON_SMALL,0,SMTO_ABORTIFHUNG,1000,h) <> 0) then begin
        result := CloneIcon(h);
        EXIT;
    end;
    Windows.SetLastError(ERROR_SUCCESS);

    if (SendMessageTimeout(owner, WM_GETICON, ICON_BIG,0,SMTO_ABORTIFHUNG,1000,h) <> 0) then begin
        result := CloneIcon(h);
        EXIT;
    end;
    Windows.SetLastError(ERROR_SUCCESS);

    frmMainPopup.AppendLog('Clone failed: ' + SysUtils.SysErrorMessage(Windows.GetLastError));
end;





//////////////////////////////////////////////////////////////
// User Interface
// close the window,
// move item
// delete item
//////////////////////////////////////////////////////////////

procedure TfrmClipboardManager.btnCloseClick(Sender: TObject);
begin
    self.ModalResult := mrok;
    self.close;
end;

procedure TfrmClipboardManager.btnTopClick(Sender: TObject);
begin
    UnitListBoxMover.MoveSelectedTop(lbQueue);
end;

procedure TfrmClipboardManager.btnUpClick(Sender: TObject);
begin
    UnitListBoxMover.MoveSelectedUp(lbQueue);
end;

procedure TfrmClipboardManager.btnDownClick(Sender: TObject);
begin
    UnitListBoxMover.MoveSelectedDown(lbQueue);
end;

procedure TfrmClipboardManager.btnBottomClick(Sender: TObject);
begin
    UnitListBoxMover.MoveSelectedBottom(lbQueue);
end;

procedure TfrmClipboardManager.btnDeleteClick(Sender: TObject);
var i : integer;
begin
    for i := (lbqueue.Items.count - 1) downto 0 do begin
        if (lbqueue.Selected[i]) then begin
            lbQueue.Items.Delete(i);
            ClipQueue.DeleteQueueItem(i);
        end;
    end;
end;


//------------------------------------------------------------------
// To set the changes in the queue, send back what we've altered
// NOTE: See the dirty trick used as a Win9X workaround
//
//------------------------------------------------------------------
procedure TfrmClipboardManager.FormClose(Sender: TObject;
  var Action: TCloseAction);
var i, index : longint;

begin
    //
    // recreate the order as seen in the form
    //
    for i := (lbQueue.Items.count - 1) downto 0 do begin
        index := ClipQueue.IndexOf( PChar(lbQueue.items.Objects[i]));
        ClipQueue.Move(index, i);
    end;
end;

procedure TfrmClipboardManager.FormShow(Sender: TObject);
var I : longint;
begin
    // dirty trick to get around Win9X memory limitations
    // for some components
    // I'm storing the full string in the object list
    lbQueue.Items.Clear;
    for i := 0 to ClipQueue.GetQueueCount -1 do begin
        lbQueue.Items.AddObject(LeftStr(ClipQueue.GetQueueItem(i),300),
                                TObject(PChar(ClipQueue.GetQueueItem(i))));
    end;
end;

//
// show the preview when an item is clicke on
//
procedure TfrmClipboardManager.lbQueueClick(Sender: TObject);
begin
    mPreview.Text := LeftStr(PChar(lbQueue.Items.Objects[lbQueue.ItemIndex]),10000);
end;

procedure TfrmClipboardManager.bCopyToClipboardClick(Sender: TObject);
var i : longint;
    s : string;
begin
    for i := 0 to (lbQueue.Items.count - 1) do begin
        if lbQueue.Selected[i] then begin
            s := s + PChar(lbQueue.items.Objects[i]);
        end;
    end;

    if (s <> '') then begin
        //self.InformOfPaste;
        self.SetIgnoreClipboard(true);
        Clipboard.SetTextBuf(PChar(s));
        self.SetIgnoreClipboard(false);
    end;
end;

end.
