unit UnitOtherQueue;

{
    Purpose:
        Create a queue of arbitrary clipboard types

    NOTES:
        Data is entered by copying the Clipboard's contents.
        Data is removed explicityly by DeleteClip or implicitly
        when the queue fills to its capacity.
        Only DeleteClip may be used to remove an item from the queue.
}
interface

uses Classes, Windows, Graphics;

type TOtherQueue = class(TObject)
    private
        qSize : longint;
        qClipsizeLimit : cardinal;
        ClipformatList : TList;
        CliphandleList : TList;
        ClipiconList : TList;
        ClipsizeList : Tlist;
        procedure AddClip(cliptype : cardinal; cliphandle : Thandle; clipicon : HICON; clipsize : cardinal);
        function DupHandle(h : Thandle; var sizeh : cardinal) : Thandle;
    public
        constructor Create();
        destructor Destroy; override;

        {queue enumeration, manipulation}

        function GetCount : longint;
        procedure SetClipboardTo(index : longint);
        procedure SetQueueSize(size : longint);
        procedure ClearQueue;
        procedure SetClipSizeLimit(size : cardinal);
        procedure DeleteClip(index : longint);

        procedure GetDIB(index : longint; pic : TPicture);
        function GetContentName(index : longint) : string;

        function GetSize(index : longint) : cardinal;
        function GetIcon(index : longint) : HICON;
        procedure ClipboardSave(icon : HICON);

end;

var OtherQueue : TOtherQueue;

implementation

uses clipbrd, unitfrmclipboardmanager, unitfrmmainpopup,  sysutils,
forms {for application object}, Dialogs;


/////////////////////
// Public Interface
/////////////////////

constructor TOtherQueue.Create;
begin
    Clipformatlist := TList.Create();
    CliphandleList := TList.Create();
    ClipsizeList := Tlist.Create();
    ClipiconList := Tlist.Create();
    qSize := 5;
    qClipsizeLimit := 500 * 1024; // 500 KB
end;
destructor TOtherQueue.Destroy;
begin
    ClipformatList.Free;
    CliphandleList.Free;
    ClipsizeList.Free;
    ClipiconList.Free;

    inherited destroy;
end;

function TOtherQueue.GetCount : longint;
begin
    result := ClipformatList.Count;
end;



function TOtherQueue.GetSize(index : longint) : cardinal;
begin
    result := Cardinal(ClipsizeList.items[index]);
end;
procedure TOtherQueue.ClearQueue;
var i : longint;
begin
    for i := (self.GetCount - 1) downto 0 do begin
        self.DeleteClip(i);
    end;
end;

function TOtherQueue.GetIcon(index : longint) : HICON;
begin
    result := HICON( ClipiconList.Items[index] );
end;

procedure TOtherQueue.GetDIB(index : longint; pic : TPicture);
var format : Cardinal;

    procedure DibToBitmap(hDIB:  THandle; BM : TBitmap);
    var
        bmfh :  TBitmapFileHeader;
        bi :  PBitmapInfo;
        ms :  TMemoryStream;
        ColorsUsed : integer;
    begin
        bi := PBitmapInfo( GlobalLock(hDIB) );
        try
            // Caculate the number of colors used (power of 2)
            ColorsUsed := bi.bmiHeader.biClrUsed;
            if (ColorsUsed = 0) and (bi.bmiHeader.biBitCount <= 8) then
                ColorsUsed := 1 shl bi.bmiHeader.biBitCount;

            bmfh.bfType := $4D42;  // 'BM'
            bmfh.bfReserved1 := 0;
            bmfh.bfReserved2 := 0;

            // point to location of actual data
            // header is variable because of the RGBQuads
            bmfh.bfOffBits := SizeOf(TBitmapFileHeader)       +
                       SizeOf(TBitmapInfoHeader)       +
                       ColorsUsed * SizeOf(TRGBQuad);
            bmfh.bfSize := bmfh.bfOffBits + bi.bmiHeader.biSizeImage;

            // Create a fake bitmap file
            // and load it into a TBitmap
            ms := TMemoryStream.Create;
            try
                ms.Write(bmfh, SizeOf(TBitmapFileHeader));
                ms.Write(bi^, bmfh.bfSize - SizeOf(TBitmapFileHeader));
                ms.Position := 0;

                BM.LoadFromStream(ms)
            finally
                ms.Free
            end;
        finally
            Windows.GlobalUnlock(hDIB);
        end;
    end;

begin
    format := cardinal(ClipformatList.Items[index]);
    if (format = CF_DIB) then begin
        DibToBitmap(Cardinal(CliphandleList.items[index]), pic.Bitmap);
    end;
end;



function TOtherQueue.GetContentName(index : longint) : string;
var name : array[0 .. 80] of char;
    format : cardinal;
    p : ^tagBITMAPINFO;
begin

    format := cardinal(ClipformatList.Items[index]);
    case (format) of
        CF_DIB          : begin
            p := GlobalLock(cardinal(CliphandleList.items[index]));
            result := 'Picture (DIB) ' + IntToStr(p^.bmiHeader.biWidth) + 'x' + IntToStr(p^.bmiHeader.biHeight);
            GlobalUnlock(cardinal(CliphandleList.items[index]));
        end;
        CF_WAVE         : result := 'Wave Audio';
        CF_BITMAP       : result := 'Picture (Bitmap)';
        CF_HDROP        : result := 'File(s)';

        CF_DIF          : result := 'CF_DIF';
        CF_TEXT         : result := 'CF_TEXT';
        CF_SYLK         : result := 'CF_SYLK';
        CF_TIFF         : result := 'CF_TIFF';
        CF_RIFF         : result := 'CF_RIFF';
        CF_LOCALE       : result := 'CF_LOCALE';
        CF_OEMTEXT      : result := 'CF_OEMTEXT';
        CF_PALETTE      : result := 'CF_PALETTE';
        CF_PENDATA      : result := 'CF_PENDATA';
        CF_UNICODETEXT  : result := 'CF_UNICODETEXT';
        CF_ENHMETAFILE  : result := 'CF_ENHMETAFILE';
        CF_METAFILEPICT : result := 'CF_METAFILEPICT';
        else begin
            GetClipboardFormatName(format,  @name,  sizeof(name)  );
            if string(name) = '' then begin
                result := 'Unknown';
            end else begin
                Result := String(name);
            end;
        end;
   end;

end;
// the ONLY place a clip may be removed from the list
procedure TOtherQueue.DeleteClip(index: longint);
begin
    //
    // free the clip memory and an associated icon
    //
    ClipformatList.Delete(index);
    Windows.GlobalFree(Cardinal(CliphandleList.Items[index]));
    CliphandleList.Delete(index);
    Windows.DestroyIcon(HICON( ClipiconList.Items[index] ));
    ClipiconList.Delete(index);
    ClipsizeList.Delete(index);
end;
procedure TOtherQueue.SetQueueSize(size : longint);
begin
    //
    // Delete the oldest items
    qSize := size;
    while (ClipformatList.Count > qSize) and (ClipformatList.Count > 0) do begin
        self.DeleteClip(0);
    end;
end;


procedure TOTherQueue.SetClipSizeLimit(size : cardinal);
begin
    qClipsizeLimit := size;
end;


////////////////////////////
// Private Implemenation
////////////////////////////

// The ONLY place a clip may be added to the list
procedure TOtherQueue.AddClip(
    cliptype : cardinal;
    cliphandle : Thandle;
    clipicon: HICON;
    clipsize : cardinal);
begin
    // flat out ignore duplicates
    // enforce queue size restrictions 
    If (CliphandleList.IndexOf(Pointer(cliphandle)) = -1) then begin
        ClipformatList.Add(Pointer(cliptype));
        CliphandleList.Add(Pointer(cliphandle));
        ClipsizeList.Add(Pointer(clipsize));
        ClipiconList.Add(Pointer(clipicon));
        Self.SetQueueSize(self.qSize);
    end;
end;


procedure TOtherQueue.ClipboardSave(icon : HICON);
var i : integer;
    h : THandle;

    clipSize : cardinal;
    ClipformatPop : word;
    CliphandlePop : THandle;
begin

    FrmMainPopup.AppendLog('<OtherQueue ClipboardSave>');


    // find the current format
    // must be non-zero
    ClipformatPop := 0;
    if not Windows.OpenClipboard(Application.Handle) then
        EXIT;

    for i := 0 to clipboard.FormatCount - 1 do begin
        if (clipboard.HasFormat(Clipboard.Formats[i])) then begin
            ClipFormatPop := Clipboard.Formats[i];
        end;
    end;
    FrmMainPopup.AppendLog('<OtherQueue ClipboardSave Format = ' + IntToStr(ClipFormatPop and $1f));

    // CF_BITMAP kills over in DupHandle, CD_DIB does not
    // A wave clip may show up as a speaker icon if CF_WAVE is not
    // before CF_DIB
    
    if clipboard.HasFormat(CF_WAVE) then
        ClipFormatPop := CF_WAVE
    else if clipboard.HasFormat(CF_DIB) then
        ClipFormatPop := CF_DIB
    else if clipboard.HasFormat(CF_HDROP) then
        ClipFormatPop := CF_HDROP;

    Windows.CloseClipboard;
    if (ClipFormatPop  = 0) then
        EXIT;


    //
    // Dupe the handle, exit on error
    //
    FrmMainPopup.AppendLog('<OtherQueue ClipboardSave duphandle>');
    h := clipboard.GetAsHandle(ClipFormatPop);
    CliphandlePop := self.DupHandle(h, ClipSize);
    if (CliphandlePop = 0) then
        EXIT;

    FrmMainPopup.AppendLog('<OtherQueue ClipboardSave size>');


    //
    // Push on lists
    //
    self.AddClip(ClipformatPop, CliphandlePop, icon, ClipSize);
    FrmMainPopup.AppendLog('<OtherQueue ClipboardSave success!>');
end;

//
// RETURNS:  Handle,
//           Zero if failed or too big
//
function TOtherQueue.DupHandle(h : Thandle; var sizeh : cardinal) : Thandle;
var
    newh : Thandle;
    p1, p2 : pointer;
begin
    result := 0;
    if (h = 0) then
        EXIT;

    //
    // make sure size is non-zero
    // and not bigger than size restrictions
    //
    sizeh := Windows.GlobalSize(h);
    if (sizeh = 0) or (sizeh > qClipsizeLimit) then
        EXIT;

    //
    // Lock and copy
    //

    newh := Windows.GlobalAlloc(GMEM_MOVEABLE, sizeh);
    p1 := Windows.GlobalLock(newh);
    p2 := Windows.GlobalLock(h);

    CopyMemory(p1, p2, sizeh);

    Windows.GlobalUnlock(h);
    Windows.GlobalUnlock(newh);

    result := newh;
end;

procedure TOtherQueue.SetClipboardTo(index : longint);
var ClipformatPop : cardinal;
    CliphandlePop : THandle;
    size : cardinal;
begin
    ClipformatPop := Cardinal(ClipformatList.Items[index]);
    CliphandlePop := Self.DupHandle( THandle(CliphandleList.Items[index]), size );

    if (ClipHandlePop <> 0) and (ClipFormatPop <> CF_TEXT) then begin
        Clipboard.SetAsHandle(ClipFormatPop, ClipHandlePop);
    end;
end;






initialization
    OtherQueue := TOtherQueue.Create;
end.
